/* 
 Lopo.java v0.2 by Peter Alexay, Jan.28.1999 Wien 
 Usage : java Lopo URL min_size max_size ext1 [ext2 ...]
*/

import java.io.*;
import java.net.*;
import java.util.*;

public class Lopo
{
  public static int allowedSizeMin,allowedSizeMax;
  public static String[] allowedExtension;
  public static String site; // site name with protocol, without '/' at end
  public static String currDir; // current whole dir name without site, with '/' at begin and end
  public static final String HttpListFileName="HTTPLIST.TXT";
  public static final String TempHTMLFilename="HTMLFILE.HTML";
  public static final String LopoListFileName="LOPOLIST.TXT";
  public static final String CRLF="\r\n";
  public static boolean DoExtFilter=false; // if true, Lopo does extension filtering
  public static boolean TestMode=false; // if false, Lopo does not download the files found
  public static boolean DeleteLopoFile=false; // if true, Lopo deletes LopoFile at start
  public static boolean IncludeImages=false; // if true, Lopo downloads image files (<IMG SRC> tag), too

  public static final int TAG_OTHER  = 0;
  public static final int TAG_ANCHOR = 1;
  public static final int TAG_IMAGE  = 2;


  public static void usage()
  {
    System.out.println("Usage : java Lopo URL min_size max_size ext1 [ext2 ...]");
    System.exit(1);
  } // usage()


  public static void writeStringInRandomFile(RandomAccessFile file,String str)
    throws IOException
  {
    StringBuffer buf=new StringBuffer(str);
    buf.append(CRLF);
    file.writeBytes(buf.toString());
  } // writeStringInRandomFile()


  public static String getFilenameFromURL(String url)
  {
    int beginIndex=url.lastIndexOf('/');
    return url.substring(beginIndex+1);
  } // getFilenameFromURL()


  public static void download_file(DataInputStream theContent,String localFilename)
    throws IOException
  {
    int theByte;
    FileOutputStream outFile=new FileOutputStream(localFilename);

//    if (len>0)
//    {
//      for (int i=0; i<len; i++)
//      {
//        theByte=theContent.read();
//        outFile.write(theByte);
//      }
//    }
//    else

    while ((theByte=theContent.read())!=-1) outFile.write(theByte);

    outFile.close();
  } // download_file()


  public static void get_file(String url)
    throws MalformedURLException, IOException
  {
    URL urlObj;
    URLConnection urlConn;
    int len,counter;
    DataInputStream theContent;
    String localFilename;

    if (checkInLopoList(url)) return; // already downloaded !

    urlObj  = new URL(url);
    urlConn = urlObj.openConnection();

    len=urlConn.getContentLength();
    localFilename=getFilenameFromURL(url);

    if (len<0)
    {
      System.out.println("get_file(): length = "+len+", skipping size check");
    }
    else if (!size_check(len)) return; // not downloadable

    if (TestMode)
    {
      System.out.println("get_file(): skipping download from URL : " + url);
      System.out.println("get_file(): Local filename : "+localFilename);
      writeInLopoList(url,localFilename);
      return; // skip download !
    } // if

    theContent=new DataInputStream(urlConn.getInputStream());

    System.out.println("get_file(): loading file from URL : " + url);
    System.out.println("get_file(): Local filename : "+localFilename);

    if (new File(localFilename).exists())
    {
      int index=localFilename.lastIndexOf('.');
      if (index==-1) // filename without '.', make new filename with '_1' at the end
      {
        counter=1;
        while (new File(localFilename+String.valueOf(counter)).exists()) counter++;
        localFilename=localFilename+String.valueOf(counter);
      }
      else // '.' exists, make new filename with '_1' before the extension
      {
        String tmp_filename=localFilename.substring(0,index); // localFilename without extension
        String tmp_ext=localFilename.substring(index); // ext. with '.'
        tmp_filename=tmp_filename+"_";
        counter=1;

        while (new File(tmp_filename+String.valueOf(counter)+tmp_ext).exists()) counter++;
        localFilename=tmp_filename+String.valueOf(counter)+tmp_ext;
      } // else
      System.out.println("get_file(): Found duplicate local file, new filename = " +
                         localFilename);
    } // if

    download_file(theContent,localFilename);
    writeInLopoList(url,localFilename);
  } // get_file()


  public static boolean size_check(int fsize)
    throws IOException
  {
    boolean result=false; // default : not downloadable

    // Extesion is already tested.

    if ((fsize>=allowedSizeMin) && (fsize<=allowedSizeMax))
    {
      result=true; // OK, downloadable
    }
    return result;
  } // size_check()


  public static void writeInLopoList(String url,String localName)
    throws IOException
  {
    // open for append :
    PrintWriter lopoListFile=new PrintWriter(new FileWriter(LopoListFileName,true));
    lopoListFile.println(url+" -> "+localName);
    lopoListFile.close();
  } // writeInLopoList()


  public static boolean checkInLopoList(String url)
    throws FileNotFoundException,IOException
  // Return-value : TRUE - url exists, FALSE - url does not exist
  {
    String line;
    BufferedReader lopoListFile=new BufferedReader(new FileReader(LopoListFileName));

    while ((line=lopoListFile.readLine())!=null)
    {
      if (line.startsWith(url)) { lopoListFile.close(); return true; } // line exists
    } // while
    lopoListFile.close();
    return false; // line does not exist
  } // checkInLopoList()


  public static void writeInHttpList(String url)
    throws IOException
  {
      // open for append :
//    PrintWriter httpListFile=new PrintWriter(new FileWriter(HttpListFileName,true));
//    httpListFile.println(url);
//    httpListFile.close();

    RandomAccessFile rfile=new RandomAccessFile(HttpListFileName,"rw");
    rfile.seek(rfile.length()); // for append

    writeStringInRandomFile(rfile,url);
    rfile.close();
  } // writeInHttpList()


  public static boolean checkInHttpList(String url)
    throws FileNotFoundException,IOException
  // Return-value : TRUE - url exists, FALSE - url does not exist
  {
    String line;
//    BufferedReader httpListFile=new BufferedReader(new FileReader(HttpListFileName));
    RandomAccessFile rfile=new RandomAccessFile(HttpListFileName,"r");

    while ((line=rfile.readLine())!=null) // attention ! line contains \r at the end !
    {
      if (line.startsWith(url)) { rfile.close(); return true; } // line exists
    } // while
    rfile.close();
    return false; // line does not exist
  } // checkInHttpList()


  public static long initialize_listfiles(String url)
    throws IOException
  {
    File checkFile=new File(HttpListFileName);
    if (checkFile.exists()) checkFile.delete(); // deletes old file ***

    // opens new file for writing
    RandomAccessFile rfile=new RandomAccessFile(HttpListFileName,"rw");

    writeStringInRandomFile(rfile,url);
    long pos=rfile.getFilePointer();
    rfile.close();
    System.out.println("HttpListFile created");

    if (DeleteLopoFile)
    {
      FileWriter lopoFile=new FileWriter(LopoListFileName,false); // creates new file
      lopoFile.close();
      System.out.println("LopoFile created");
    }

    return pos;
  } // initialize_listfiles()


  public static boolean get_tag(BufferedReader stream, StringBuffer tag)
    throws IOException
  { // return value : true-tag found, false-eof
    char theByte;

    tag.setLength(0);

    while ((theByte=(char)(stream.read())) != (char)-1)
    {
      if (theByte=='<')
      {
        while ((theByte=(char)(stream.read())) != '>')
          tag.append(theByte);
        return true; // tag found
      }
    } // while
    return false; // eof
  } // get_tag()


  public static void updateCurrDir(String url)
  // Makes currDir : current whole dir name without site, with '/' at begin and end
  // (If the current url is a directory, it must have a '/' at the end !)
  {
    int perperIndex=url.indexOf("//");
    int perIndex=url.indexOf("/",perperIndex+2); // first '/' in the url
    if (perIndex==-1)
    {
      url=url+"/";
      perIndex=url.indexOf("/",perperIndex+2);
    }
    int lastperIndex=url.lastIndexOf("/");
    if (perIndex==lastperIndex) currDir="/"; // the root dir.
    else currDir=url.substring(perIndex,lastperIndex+1);
  } // updateCurrDir()


  public static void download_tempfile(String url)
    throws MalformedURLException,IOException
  // Downloads a html textfile into a local temporary file
  {
    URL urlObj=new URL(url);
    URLConnection urlConn=urlObj.openConnection();
    BufferedReader theContent=new BufferedReader(new InputStreamReader(urlConn.getInputStream()));
    FileWriter writer=new FileWriter(TempHTMLFilename);
    String line;

    System.out.println("download_tempfile(): downloading HTML-file from URL : " + url);

    while ((line=theContent.readLine()) != null)
    {
      writer.write(line+CRLF);
    } // while

    writer.close();

//    download_file(theContent,TempHTMLFilename);
  } // download_tempfile()


  public static boolean isHtmlFile(String url)
    throws MalformedURLException,IOException
  {
    URL urlObj=new URL(url);
    URLConnection urlConn=urlObj.openConnection();
    String contentType;

    contentType=urlConn.getContentType();
    return (contentType.equals("text/html"));
  } // isHtmlFile()

  public static boolean isHtmlAtEnd(String url)
  {
    // s5="/";
    return (url.endsWith("HTM")  || url.endsWith("htm") ||
            url.endsWith("HTML") || url.endsWith("html"));
  } // isHtmlAtEnd()


  public static boolean extensionCheck(String url)
  {
    for (int i=0; i<allowedExtension.length; i++)
    {
      if (url.endsWith(allowedExtension[i])) return true; // allowed
    }
    return false; // not allowed
  } // extensionCheck()


  public static String removeUpperDir(String url)
  {
    int upperInd,perInd;
    String firstStr,secondStr;

    while ((upperInd=url.indexOf("../"))!=-1)
    {
      secondStr=url.substring(upperInd+3); // the end of string, without / at begin
      firstStr=url.substring(0,upperInd-1);
      perInd=firstStr.lastIndexOf("/");
      firstStr=firstStr.substring(0,perInd+1); // with / at end
      url=firstStr+secondStr;
    } // while

    return url;
  } // removeUpperDir()


  public static void parse_html(String url)
    throws MalformedURLException,IOException
  {
    int hrefIndex,linkIndex;
    String link,linkToParse;
    boolean allowed; // flag for same or other server
    long position=0L;
    boolean end_of_list=false,file_OK;

    position=initialize_listfiles(url); // Writes url into HttpList as first element
    updateCurrDir(url);
    
    do
    {
      BufferedReader htmlfile=new BufferedReader(new FileReader(TempHTMLFilename));
      StringBuffer tag=new StringBuffer();
      
      while (get_tag(htmlfile,tag))
      {
        String tagStr=tag.toString();
        int tagType=TAG_OTHER;

        if (tagStr.startsWith("A ")  || tagStr.startsWith("a ") ||
            tagStr.startsWith("A\n") || tagStr.startsWith("a\n")||
            tagStr.startsWith("A\r") || tagStr.startsWith("a\r")) // check for anchor
          tagType=TAG_ANCHOR;
        else if (tagStr.startsWith("IMG ")  || tagStr.startsWith("img ") ||
                 tagStr.startsWith("IMG\n") || tagStr.startsWith("img\n")||
                 tagStr.startsWith("IMG\r") || tagStr.startsWith("img\r")) // check for image
          tagType=TAG_IMAGE;
        
        if ((tagType==TAG_ANCHOR) || ((tagType==TAG_IMAGE) && IncludeImages))
        {
          if (tagType==TAG_ANCHOR) hrefIndex=tagStr.indexOf("HREF");
          else hrefIndex=tagStr.indexOf("SRC");
          if (hrefIndex==-1)
          {
            if (tagType==TAG_ANCHOR) hrefIndex=tagStr.indexOf("href");
            else hrefIndex=tagStr.indexOf("src");
          }
          
          if (hrefIndex!=-1) // link found
          {
            linkIndex=tagStr.indexOf('"',hrefIndex);
            if (linkIndex==-1) // no " in href
            {
              linkIndex=tagStr.indexOf('=',hrefIndex)+1;
              while (tagStr.charAt(linkIndex)==' ') linkIndex++; // skip spaces
              int endIndex=linkIndex;
              while ((endIndex<tagStr.length()) &&
                    ((tagStr.charAt(endIndex)!='>') &&
                     (tagStr.charAt(endIndex)!=' ') &&
                     (tagStr.charAt(endIndex)!='\t') &&
                     (tagStr.charAt(endIndex)!='"')))
                endIndex++;
              if (endIndex==tagStr.length())
                link=tagStr.substring(linkIndex); // to EOS
              else
                link=tagStr.substring(linkIndex,endIndex);
            }
            else // link between "s
            {
              link=tagStr.substring(linkIndex+1,tagStr.indexOf('"',linkIndex+1));
            }
            if (TestMode) System.out.println("parse_html(): Found link : "+link);
            allowed=true; // default : on the same server
            
            // Check link : local oder internet
            if (link.indexOf("//")==-1) // local
            {
              // Check path (link) : absolute oder relative
              if (link.startsWith("/")) // absolute
              {
                linkToParse=site+link;
              }
              else // relative
              {
                linkToParse=site+currDir+link;
              } // else
            }
            else // internet-address
            {
              int perperIndex=link.indexOf("//");
              int perIndex=link.indexOf("/",perperIndex+2);
              if (perIndex==-1) // no / at the end, append one !
              {
                link=link+"/";
                perIndex=link.indexOf("/",perperIndex+2);
              }
              String newServer=link.substring(0,perIndex);
              // Check server : the same or other
              if (newServer.equals(site)) // OK, the same server
              {
                linkToParse=link;
              }
              else // other server -> not allowed
              {
                allowed=false; // on other server
                linkToParse=null;
              } // else
            } // else
            
            if (allowed) // only link on the same server allowed
            {
              System.out.println("Link to parse : "+linkToParse);
              
              linkToParse=removeUpperDir(linkToParse);
              if (isHtmlFile(linkToParse))
              {
                if (!isHtmlAtEnd(linkToParse)) // location is a directory
                {
                  if (linkToParse.charAt(linkToParse.length()-1)!='/') // the last character != '/'
                    linkToParse=linkToParse+"/"; // append '/' to the end
                }

                if (!checkInHttpList(linkToParse)) // not yet parsed
                {
                  writeInHttpList(linkToParse); // append to HttpList
                  if (TestMode) System.out.println("Written in HttpList");
                }
              }
              
              if (DoExtFilter) // do extension filtering ?
              {
                // Check linkToParse for file-type, if good -> download it
                if (extensionCheck(linkToParse))
                {
                  get_file(linkToParse); // download file, if it was not downloaded yet
                } // if
              }
              else get_file(linkToParse); // always download file
            } // if (allowed)
            else System.out.println("parse_html(): Link not allowed to download : "+link);
          } // if "href" found in anchor
        } // if anchor or image found in tagStr
      } // while (get_tag(...))
      
      htmlfile.close();
      
      do
      {
        file_OK=false;
        RandomAccessFile rfile=new RandomAccessFile(HttpListFileName,"r");
        try {
          rfile.seek(position);
        } catch (IOException e) { end_of_list=true; }

        if (!end_of_list)
        {
          String location=rfile.readLine();
          if (TestMode) System.out.println("parse_html(): Next location : "+location);
          position=rfile.getFilePointer();
          rfile.close();

          if (location==null) end_of_list=true;
          else
          {
            if (location.endsWith("\r"))
            {
              location=location.substring(0,location.length()-1);
            }
            if (isHtmlAtEnd(location) || location.endsWith("/")) // html file or directory
            {
              if (TestMode) System.out.println(
                  "parse_html(): Is a HTML file or directory, downloading and parsing ...");
              download_tempfile(location);
              updateCurrDir(location); // a '/' is already appended if it was a directory
              file_OK=true;
            } // if
          } // else
        } // if
      } while (!end_of_list && !file_OK);

    } while (!end_of_list);
  } // parse_html()


  public static boolean getYNfromKeyboard(String str)
    throws IOException
  {
    BufferedReader keyb=new BufferedReader(new InputStreamReader(System.in));
    String inputBuf;
    boolean inputOk=false,result=false;

    do
    {
      System.out.print(str); System.out.flush();
      inputBuf=keyb.readLine();
      if (inputBuf.equalsIgnoreCase("Y") || inputBuf.equalsIgnoreCase("YES"))
      {
        result=true; inputOk=true;
      }
      else if (inputBuf.equalsIgnoreCase("N") || inputBuf.equalsIgnoreCase("NO"))
      {
        result=false; inputOk=true;
      }
    } while (!inputOk);

    return result;
  } // getYNfromKeyboard()

  //************************************************************************************

  public static void main(String[] args)
    throws IOException
  {   // Usage : Lopo URL min_size max_size [ext1 [ext2 ...]]

    System.out.println("Lopo v0.2 by Peter Alexay, Jan.28.1999 Wien\n");
    if (args.length < 3) usage();
    String location=args[0];

    allowedSizeMin=Integer.parseInt(args[1]);
    allowedSizeMax=Integer.parseInt(args[2]);
    
    if (args.length < 4) DoExtFilter=false;
    else
    {
      DoExtFilter=true;
      allowedExtension=new String[args.length-3];
      for (int i=0; i<(args.length-3); i++) allowedExtension[i]=args[i+3];
    }

    int perperIndex=location.indexOf("//");
    if (perperIndex==-1) usage();
    int perIndex=location.indexOf('/',perperIndex+2);
    if (perIndex==-1) // there isn't any '/' in the given URL
    {
      location=location+"/";
      perIndex=location.indexOf('/',perperIndex+2);
    }

    // make site : site name with protocol, without '/' at end
    site=location.substring(0,perIndex);

    // Get TestMode, DeleteLopoFile and IncludeImages flags from the keyboard :
    TestMode=getYNfromKeyboard("Testmode (Y/N) ?");
    DeleteLopoFile=getYNfromKeyboard("Create new Lopo-file before start (Y/N) ?");
    IncludeImages=getYNfromKeyboard("Include image-files (Y/N) ?");

    if (isHtmlFile(location)) // content-type tag = "text/html" ?
    { // First arg : a html file to parse
      if (!isHtmlAtEnd(location)) // location is a directory
      {
        if (location.charAt(location.length()-1)!='/') // the last character != '/'
          location=location+"/"; // append '/' to the end
      }
      // make currDir : current whole dir name without site, with '/' at begin and end
      updateCurrDir(location);
      download_tempfile(location);
      parse_html(location);
    }
    else // nothing to parse. First arg : a file to download
    {
      long position=initialize_listfiles(location); // Writes url into HttpList as first element
      try {
        get_file(location);
      } catch (MalformedURLException e){ System.err.println("Error : Malformed URL"); }
        catch (IOException e)          { System.err.println("Error : Download not successful"); }
    } // else

    System.out.println("main(): OK");
  } // main()

} // class Lopo
